/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

Application
    conjugateFoam

Description
    Solves a simple Laplace equation, e.g. for thermal diffusion in a solid.

\*---------------------------------------------------------------------------*/

#include "fvCFD.H"
#include "Switch.H"
#include "fixedGradientFvPatchFields.H"
#include "tractionDisplacement/tractionDisplacementFvPatchVectorField.H"
#include "fvMesh.H"
// #include "typeInfo.H"
#include "PrimitivePatchInterpolation.H"
// #include "fvMesh.H"
// #include "motionSolver.H"
// #include "FixedValuePointPatchField.H"
#include "tetDecompositionMotionSolver.H"
#include "laplaceTetDecompositionMotionSolver.H"
#include "pseudoSolidTetDecompositionMotionSolver.H"

// typedef FixedValuePointPatchField<tetPolyPatchField,tetPolyPatchFaceDecomp,vector> fixedMotionPatch;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

int main(int argc, char *argv[])
{

#   include "setRootCase.H"

#   include "createTime.H"
#   include "createMeshes.H"

#   include "readMechanicalProperties.H"
#   include "readThermalProperties.H"
#   include "readCoupling.H"

#   include "createIcoFields.H"
#   include "createStructureFields.H"

#   include "createMeshMotion.H"

#   include "initContinuityErrs.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

    for (runTime++; !runTime.end(); runTime++)
    {
        Info<< "\n Time = " << runTime.timeName() << nl << endl;

	if(runTime.value()>=startMeshMotion.value()) {
	  Info << "\nMoving mesh\n" << endl;

	  dispVals=inter.faceToPointInterpolate(displace);

	  //	  tetPointVectorField neu=inter.faceToPointInterpolate(patchU);
	  scalar factor=1/(runTime.deltaT().value()*motionRelaxation.value());
	  scalar maxDist=-1e10;
	  scalar maxAway=-1e10;

	  forAll(fluidMesh.localPoints(),fluidI) {
	    label solidI=pExchange[fluidI];

	    vector here=fluidMesh.localPoints()[fluidI];
	    vector old =oldPoints[fluidI];
	    vector disp=dispVals[solidI];
	    vector neu=disp+old;
	    vector move=factor*(neu-here);
	    vector prev=patchU[fluidI];

	    //	    Info << "Punkt: " << fluidI << " \t " << solidI << " \t " << here <<  " \t " << old <<  " \t " << disp <<  " \t " << move  <<  " \t " << prev <<  " \t " << mag(here-old) << endl;

	    patchU[fluidI]=move;

	    if(mag(here-old)>maxDist) {
	      maxDist=mag(here-old);
	    }
	    if(mag(neu-here)>maxAway) {
	      maxAway=mag(here-neu);
	    }
	  }

	  mesh1.movePoints(motionPtr->newPoints());
	
	  Info << "\nBiggest movement: " << maxDist << " Bigges divergence " << maxAway <<  endl;
	}

	Info << "Solving flow in mesh1\n" << endl;
	{
#         include "readPISOControls.H"
#         include "CourantNo.H"

	  fvVectorMatrix UEqn
	    (
	     fvm::ddt(U)
	     + fvm::div(phi, U)
	     - fvm::laplacian(nu, U)
	     );

	  solve(UEqn == -fvc::grad(p));

	  // --- PISO loop

	  for (int corr=0; corr<nCorr; corr++)
	    {
	      volScalarField rUA = 1.0/UEqn.A();

	      U = rUA*UEqn.H();
	      phi = (fvc::interpolate(U) & mesh1.Sf()) 
                + fvc::ddtPhiCorr(rUA, U, phi);

	      adjustPhi(phi, U, p);

	      for (int nonOrth=0; nonOrth<=nNonOrthCorr; nonOrth++)
		{
		  fvScalarMatrix pEqn
		    (
		     fvm::laplacian(rUA, p) == fvc::div(phi)
		     );

		  pEqn.setReference(pRefCell, pRefValue);
		  pEqn.solve();

		  if (nonOrth == nNonOrthCorr)
		    {
		      phi -= pEqn.flux();
		    }
		}

#             include "continuityErrs.H"

	      U -= rUA*fvc::grad(p);
	      U.correctBoundaryConditions();
	    }

	}

	Info << "\nSolving structure in mesh2\n" << endl;

	{
#         include "readStressedFoamControls.H"
	  int iCorr = 0;
	  scalar initialResidual = 0;

	  do
	    {
	      volTensorField gradU = fvc::grad(Udisp);

	      if (thermalStress)
		{
		  solve
		    (
		     fvm::ddt(T) == fvm::laplacian(DT, T)
		     );
		}

	      fvVectorMatrix UEqn
		(
		 fvm::d2dt2(Udisp)
		 ==
		 fvm::laplacian(2*mu + lambda, Udisp, "laplacian(DUdisp,Udisp)")

		 + fvc::div
		 (
		  mu*gradU.T() + lambda*(I*tr(gradU)) - (mu + lambda)*gradU,
		  "div(sigma)"
		  )
		 );

	      if (thermalStress)
		{
		  UEqn += threeKalpha*fvc::grad(T);
		}

	      //UEqn.setComponentReference(1, 0, vector::X, 0);
	      //UEqn.setComponentReference(1, 0, vector::Z, 0);

	      initialResidual = UEqn.solve().initialResidual();

	    } while (initialResidual > convergenceTolerance && ++iCorr < nCorr);

#       include "calculateStress.H"
 
	  Info << "\nMaximum Displacement: " << max(mag(Udisp)).value() << endl;

	  
	} 

	Info << "\nCoupling the solutions\n" << endl;

	scalarField & fluidP = p.boundaryField()[fluidSideI];
	
 	scalarField &solidP = displace.pressure();
	
	//	Info << "Pressure solid before : [" << min(solidP) << " , " << max(solidP) <<"]" << endl;

	//	solidP=fluidP; // -min(fluidP);
	forAll(fluidP,fI) {
	  solidP[exchange[fI]]=fluidP[fI];
	}

	//	Info << "Pressure solid after : [" << min(solidP) << " , " << max(solidP) <<"]" << endl;

#       include "write.H"
    }

    Info<< "\n end \n";

    return(0);
}


// ************************************************************************* //
